# Copyright (C) 2014 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS).

# This file is part of mirakuru.

# mirakuru is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# mirakuru is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with mirakuru.  If not, see <http://www.gnu.org/licenses/>.
"""Mirakuru's main module."""

__version__ = '0.4.0'

from mirakuru.base import Executor
from mirakuru.output import OutputExecutor
from mirakuru.tcp import TCPExecutor
from mirakuru.http import HTTPExecutor
from mirakuru.pid import PidExecutor

from mirakuru.exceptions import (
    TimeoutExpired,
    AlreadyRunning,
)

__all__ = (
    'Executor',
    'OutputExecutor',
    'TCPExecutor',
    'HTTPExecutor',
    'PidExecutor',
    'TimeoutExpired',
    'AlreadyRunning',
)



# Copyright (C) 2014 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS).

# This file is part of mirakuru.

# mirakuru is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# mirakuru is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with mirakuru.  If not, see <http://www.gnu.org/licenses/>.
"""HTTP enabled process executor."""

import socket

from mirakuru.compat import HTTPConnection, HTTPException, OK
from mirakuru.compat import urlparse

from mirakuru.tcp import TCPExecutor


class HTTPExecutor(TCPExecutor):

    """Http enabled process executor."""

    def __init__(self, command, url, **kwargs):
        """
        Initialize HTTPExecutor executor.

        :param (str, list) command: command to run to start service
        :param str url: url where executor can check
            if process has already started.
        :param bool shell: see `subprocess.Popen`
        :param int timeout: time to wait for process to start or stop.
            if None, wait indefinitely.
        :param float sleep: how often to check for start/stop condition
        :param int sig_stop: signal used to stop process run by executor.
            default is SIGTERM
        :param int sig_kill: signal used to kill process run by  executor.
            default is SIGKILL
        """
        self.url = urlparse(url)
        """
        An :func:`urlparse.urlparse` representation of an url.

        It'll be used to check process status on."""

        super(HTTPExecutor, self).__init__(
            command, host=self.url.hostname, port=self.url.port, **kwargs
        )

    def after_start_check(self):
        """Check if defined url returns successful head."""
        try:
            conn = HTTPConnection(self.url.hostname, self.url.port)

            conn.request('HEAD', self.url.path)
            response = conn.getresponse()

            if response.status is OK:
                conn.close()
                return True

        except (HTTPException, socket.timeout, socket.error):
            return False



"""
Package of tests for mirakuru.

Tests are written using py.test framework which dictates patterns that should
be followed in test cases.
"""

from os import path


tests_path = path.abspath(path.dirname(__file__))

test_server_path = path.join(tests_path, "test_server.py")



"""
Conftest responsible for filtering ResourceWarning for python 3.

Also includes fixture for raising an error whenever we leave any resource open.
"""
import sys
from warnings import simplefilter

import pytest


if sys.version_info == '3':
    from exteptions import ResourceWarning
    simplefilter('default', ResourceWarning)


@pytest.fixture(autouse=True)
def error_warn(request, recwarn):
    """Raise error whenever any warning gets listed."""
    def list_errors_and_error():
        if recwarn.list:
            raise recwarn.list[0].message

    request.addfinalizer(list_errors_and_error)



"""
HTTP server that responses with delays used for tests.

Example usage:

    python tests/slow_server.py [host:port]

        - run HTTP Server, host and port are optional

    python tests/slow_server.py [host:port] True

        - run immortal server (stopping procces only by SIGKILL)

"""

import sys
import time
import signal

from mirakuru.compat import HTTPServer, BaseHTTPRequestHandler


class SlowServerHandler(BaseHTTPRequestHandler):

    """Slow server handler."""

    timeout = 2
    endtime = None

    def do_GET(self):
        """Serve GET request."""
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        self.wfile.write('Hi. I am very slow.')
        return

    def do_HEAD(self):
        """
        Serve HEAD request.

        but count to wait and return 500 response if wait time not exceeded
        due to the fact, that HTTPServer will hang waiting for response
        to return otherwise if none response will be returned.
        """
        if self.count_timeout():
            self.send_response(200)
        else:
            self.send_response(500)
        self.end_headers()
        return

    def count_timeout(self):
        """Count down the timeout time."""
        if SlowServerHandler.endtime is None:
            SlowServerHandler.endtime = time.time() + SlowServerHandler.timeout
        return time.time() >= SlowServerHandler.endtime


def block_signals():
    """
    Catch all of the signals that it is possible.

    Reject their default behaviour. The process is actually mortal but the
    only way to kill is to send SIGKILL signal (kill -9).
    """
    def sighandler(signum, _):
        """Signal handling function."""
        print('Tried to kill with signal {}.'.format(signum))

    for sgn in [x for x in dir(signal) if x.startswith("SIG")]:
        try:
            signum = getattr(signal, sgn)
            signal.signal(signum, sighandler)
        except (ValueError, RuntimeError, OSError):
            pass


if __name__ == "__main__":

    host, port, immortal = "127.0.0.1", 8000, False
    if len(sys.argv) in (2, 3):
        host, port = sys.argv[1].split(":")

    if len(sys.argv) == 3:
        immortal = sys.argv[2]

    if immortal:
        block_signals()

    server = HTTPServer((host, int(port)), SlowServerHandler)
    print("Starting slow server on {}:{}...".format(host, port))
    server.serve_forever()



# Copyright (C) 2014 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS).

# This file is part of mirakuru.

# mirakuru is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# mirakuru is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with mirakuru.  If not, see <http://www.gnu.org/licenses/>.
"""TCP executor definition."""

import socket
from mirakuru.base import StartCheckExecutor


class TCPExecutor(StartCheckExecutor):

    """
    TCP-listening process executor.

    Used to start (and wait to actually be running) processes that can accept
    TCP connections.
    """

    def __init__(self, command, host, port, **kwargs):
        """
        Initialize TCPExecutor executor.

        :param (str, list) command: command to run to start service
        :param str host: host under which process is accessible
        :param int port: port under which process is accessible
        :param bool shell: see `subprocess.Popen`
        :param int timeout: time to wait for process to start or stop.
            if None, wait indefinitely.
        :param float sleep: how often to check for start/stop condition
        :param int sig_stop: signal used to stop process run by executor.
            default is SIGTERM
        :param int sig_kill: signal used to kill process run by  executor.
            default is SIGKILL
        """
        super(TCPExecutor, self).__init__(command, **kwargs)
        self.host = host
        """Host name, process is listening on."""
        self.port = port
        """Port number, process is listening on."""

    def pre_start_check(self):
        """
        Check if process accepts connections.

        .. note::

            Process will be considered started, when it'll be able to accept
            TCP connections as defined in initializer.
        """
        try:
            sock = socket.socket()
            sock.connect((self.host, self.port))
            return True
        except (socket.error, socket.timeout):
            return False
        finally:
            # close socket manually for sake of pypy
            sock.close()

    def after_start_check(self):
        """
        Check if process accepts connections.

        .. note::

            Process will be considered started, when it'll be able to accept
            TCP connections as defined in initializer.
        """
        return self.pre_start_check()  # we can reuse logic from `pre_start()`



# Copyright (C) 2014 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS).

# This file is part of mirakuru.

# mirakuru is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# mirakuru is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with mirakuru.  If not, see <http://www.gnu.org/licenses/>.
"""mirakuru's python2 and python3 compatibility module."""

import sys


python = sys.executable

if sys.version_info.major == 2:
    from httplib import HTTPConnection, HTTPException, OK
    from BaseHTTPServer import HTTPServer, BaseHTTPRequestHandler
    from urlparse import urlparse
    http_server_cmd = "{python} -m SimpleHTTPServer".format(python=python)
else:
    # In python3 httplib is renamed to http.client
    from http.client import HTTPConnection, HTTPException, OK
    # In python3 BaseHTTPServer is renamed to http.server
    from http.server import HTTPServer, BaseHTTPRequestHandler
    # In python3 urlparse is renamed to urllib.parse
    from urllib.parse import urlparse
    http_server_cmd = "{python} -m http.server".format(python=python)


__all__ = (
    'HTTPConnection',
    'HTTPException',
    'OK',
    'HTTPServer',
    'BaseHTTPRequestHandler',
    'urlparse',
    'http_server_cmd',
)



# Copyright (C) 2014 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS).

# This file is part of mirakuru.

# mirakuru is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# mirakuru is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with mirakuru.  If not, see <http://www.gnu.org/licenses/>.
"""This executor awaits for appearance of a predefined banner in output."""

import re
import select

from mirakuru.base import Executor


class OutputExecutor(Executor):

    """Executor that awaits for string output being present in output."""

    def __init__(self, command, banner, **kwargs):
        """
        Initialize OutputExecutor executor.

        :param (str, list) command: command to run to start service
        :param str banner: string that has to appear in process output -
            should compile to regular expression.
        :param bool shell: see `subprocess.Popen`
        :param int timeout: time to wait for process to start or stop.
            if None, wait indefinitely.
        :param float sleep: how often to check for start/stop condition
        :param int sig_stop: signal used to stop process run by executor.
            default is SIGTERM
        :param int sig_kill: signal used to kill process run by  executor.
            default is SIGKILL
        """
        super(OutputExecutor, self).__init__(command, **kwargs)
        self._banner = re.compile(banner)
        self.poll_obj = None

    def start(self):
        """
        Start process.

        .. note::

            Process will be considered started, when defined banner will appear
            in process output.
        """
        super(OutputExecutor, self).start()

        # get a polling object
        self.poll_obj = select.poll()

        # register a file descriptor
        # POLLIN because we will wait for data to read
        self.poll_obj.register(self.output(), select.POLLIN)

        try:
            self.wait_for(self._wait_for_output)

            # unregister the file descriptor and delete the polling object
            self.poll_obj.unregister(self.output())
        finally:
            del self.poll_obj

    def _wait_for_output(self):
        """
        Check if output matches banner.

        .. warning::
            Waiting for I/O completion. It does not work on Windows. Sorry.
        """
        # Here we should get an empty list or list with a tuple [(fd, event)]
        # When we get list with a tuple we can use readline method on
        # the file descriptor.
        poll_result = self.poll_obj.poll(0)

        if poll_result:
            line = self.output().readline()
            if self._banner.match(line):
                return True

        return False



"""Pid Executor tests."""
import os

import pytest

from mirakuru import PidExecutor
from mirakuru import TimeoutExpired, AlreadyRunning


filename = "pid-test-tmp{0}".format(os.getpid())
process = 'bash -c "sleep 2 && touch {0}"'.format(filename)


@pytest.yield_fixture(autouse=True)
def run_around_tests():
    """
    Make sure the **filename** file is not present.

    This executor actually removes filename as process used to test
    PidExecutor only creates it.
    """
    try:
        os.remove(filename)
    except OSError:
        pass

    yield

    try:
        os.remove(filename)
    except OSError:
        pass


@pytest.mark.parametrize('timeout', (None, 5))
def test_start_and_wait(timeout):
    """Test if the executor will await for the process to create a file."""
    process = 'bash -c "sleep 2 && touch {0}  && sleep 10"'.format(filename)
    executor = PidExecutor(process, filename, timeout=timeout)
    with executor:
        assert executor.running() is True

    # check proper __str__ and __repr__ rendering:
    assert 'PidExecutor' in repr(executor)
    assert process in str(executor)


@pytest.mark.parametrize('pid_file', (None, ""))
def test_empty_filename(pid_file):
    """Check whether an exception is raised if an empty filename is given."""
    with pytest.raises(ValueError):
        PidExecutor(process, pid_file)


def test_if_file_created():
    """Check whether the process really created the given file."""
    assert os.path.isfile(filename) is False
    executor = PidExecutor(process, filename)
    with executor:
        assert os.path.isfile(filename) is True


def test_timeout_error():
    """Check if timeout properly expires."""
    executor = PidExecutor(process, filename, timeout=1)

    with pytest.raises(TimeoutExpired):
        executor.start()

    assert executor.running() is False


def test_fail_if_other_executor_running():
    """Test raising AlreadyRunning exception when port is blocked."""
    process = 'bash -c "sleep 2 && touch {0}  && sleep 10"'.format(filename)
    executor = PidExecutor(process, filename)
    executor2 = PidExecutor(process, filename)

    with executor:

        assert executor.running() is True

        with pytest.raises(AlreadyRunning):
            executor2.start()



"""General mirakuru library tests."""
from mirakuru import *  # noqa


def test_importing_mirakuru():
    """Test if all most commonly used classes are imported by default."""
    assert HTTPExecutor
    assert TCPExecutor
    assert Executor
    assert OutputExecutor
    assert TCPExecutor
    assert HTTPExecutor
    assert PidExecutor
    assert TimeoutExpired
    assert AlreadyRunning



"""Output executor test."""
import pytest

from mirakuru import OutputExecutor
from mirakuru.exceptions import TimeoutExpired


def test_executor_waits_for_process_output():
    """Check if executor waits for specified output."""
    command = 'bash -c "sleep 2 && echo foo && echo bar && sleep 100"'
    executor = OutputExecutor(command, 'foo')
    executor.start()

    assert executor.running() is True
    # foo has been used for start as a banner.
    assert executor.output().readline() == 'bar\n'
    executor.stop()

    # check proper __str__ and __repr__ rendering:
    assert 'OutputExecutor' in repr(executor)
    assert 'foo' in str(executor)


def test_executor_dont_start():
    """Executor should not start."""
    command = 'bash -c "sleep 2 && echo foo && echo bar && sleep 100"'
    executor = OutputExecutor(command, 'foobar', timeout=3)
    with pytest.raises(TimeoutExpired):
        executor.start()

    assert executor.running() is False



# Copyright (C) 2014 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS).

# This file is part of mirakuru.

# mirakuru is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# mirakuru is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with mirakuru.  If not, see <http://www.gnu.org/licenses/>.
"""Pid executor definition."""

import os.path
from mirakuru.base import StartCheckExecutor


class PidExecutor(StartCheckExecutor):

    """
    File existence checking process executor.

    Used to start processes that create pid files (or any other for that
    matter). Starts the given process and waits for the given file to be
    created.
    """

    def __init__(self, command, filename, **kwargs):
        """
        Initialize the PidExecutor executor.

        If the filename is empty, a ValueError is thrown.

        :param (str, list) command: command to run to start service
        :param str filename: the file which is to exist
        :param bool shell: see `subprocess.Popen`
        :param int timeout: time to wait for the process to start or stop.
            if None, wait indefinitely.
        :param float sleep: how often to check for start/stop conditions
        :param int sig_stop: signal used to stop process run by executor.
            default is SIGTERM
        :param int sig_kill: signal used to kill process run by  executor.
            default is SIGKILL

        :raises: ValueError
        """
        if not filename:
            raise ValueError("filename cannot be empty")
        super(PidExecutor, self).__init__(command, **kwargs)
        self.filename = filename
        """the name of the file which the process is to create."""

    def pre_start_check(self):
        """
        Check if the specified file has been created.

        .. note::

            The process will be considered started when it will have created
            the specified file as defined in the initializer.
        """
        return os.path.isfile(self.filename)

    def after_start_check(self):
        """
        Check if the process has created the specified file.

        .. note::

            The process will be considered started when it will have created
            the specified file as defined in the initializer.
        """
        return self.pre_start_check()  # we can reuse logic from `pre_start()`



"""Mirakuru's exceptions."""


class TimeoutExpired(Exception):

    """Is raised when the timeout expires while starting an executor."""

    def __init__(self, executor, timeout):
        """
        Exception initialization.

        :param mirakuru.base.Executor executor: for which exception occured.
        :param int timeout: timeout for which exception occurred.
        """
        super(self.__class__, self).__init__()
        self.executor = executor
        self.timeout = timeout

    def __str__(self):
        """
        Return Exception's string representation.

        :returns: string representation
        :rtype: str
        """
        return 'Executor {0} timed out after {1} seconds'.format(
            self.executor, self.timeout
        )


class AlreadyRunning(Exception):

    """
    Is raised when the executor seems to be already running.

    When some other process (not necessary executor) seems to be started with
    same configuration we can't bind to same port.
    """

    def __init__(self, executor):
        """
        Exception initialization.

        :param mirakuru.base.Executor executor: for which exception occured.
        """
        super(self.__class__, self).__init__()
        self.executor = executor

    def __str__(self):
        """
        Return Exception's string representation.

        :returns: string representation
        :rtype: str
        """
        return ("Executor {exc.executor} seems to be already running. "
                "It looks like the previous executor process hasn't been "
                "terminated or killed. Also there might be some completely "
                "different service listening on {exc.executor.port} port."
                .format(exc=self))



# Copyright (C) 2014 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS).

# This file is part of mirakuru.

# mirakuru is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# mirakuru is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with mirakuru.  If not, see <http://www.gnu.org/licenses/>.
"""Base executor with the most basic functionality."""

import os
import time
import shlex
import signal
import subprocess
from contextlib import contextmanager

from mirakuru.exceptions import TimeoutExpired, AlreadyRunning


class Executor(object):

    """Basic executor with the most basic functionality."""

    def __init__(
            self, command, shell=False, timeout=None, sleep=0.1,
            sig_stop=signal.SIGTERM, sig_kill=signal.SIGKILL
    ):
        """
        Initialize executor.

        :param (str, list) command: command to run to start service
        :param bool shell: see `subprocess.Popen`
        :param int timeout: time to wait for process to start or stop.
            if None, wait indefinitely.
        :param float sleep: how often to check for start/stop condition
        :param int sig_stop: signal used to stop process run by executor.
            default is SIGTERM
        :param int sig_kill: signal used to kill process run by  executor.
            default is SIGKILL

        .. note::

            **timeout** set for executor is valid for all the level of waits
            on the way up. That means that if some more advanced executor sets
            timout to 10 seconds, and it'll take 5 seconds for first check,
            second check will only have 5 seconds left.

        """
        if isinstance(command, (list, tuple)):
            self.command = ' '.join(command)
            """Command that executor runs."""
            self.command_parts = command
        else:
            self.command = command
            self.command_parts = shlex.split(command)

        self._shell = shell
        self._timeout = timeout
        self._sleep = sleep
        self._sig_stop = sig_stop
        self._sig_kill = sig_kill

        self._endtime = None
        self.process = None
        """A :class:`subprocess.Popen` instance once process is started."""

    def __enter__(self):
        """Enter context manager."""
        self.start()

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit context manager."""
        self.stop()

    def running(self):
        """
        Check if executor is running.

        :returns: True if process is running, False otherwise
        :rtype: bool
        """
        if self.process is None:
            return False
        return self.process.poll() is None

    def start(self):
        """
        Start defined process.

        After process gets started, timeout countdown begins as well.

        .. note::
            We want to open ``stdin``, ``stdout`` and ``stderr`` as text
            streams in universal newlines mode, so we have to set
            ``universal_newlines`` to ``True``.
        """
        if self.process is None:
            command = self.command
            if not self._shell:
                command = self.command_parts

            self.process = subprocess.Popen(
                command,
                shell=self._shell,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                universal_newlines=True,
                preexec_fn=os.setsid
            )

        self._set_timeout()

    def _set_timeout(self, timeout=None):
        """
        Set timout for possible wait.

        :param int timeout: [optional] specific timeout to set.
            If not set, Executor._timeout will be used instead.
        """
        timeout = timeout or self._timeout

        if timeout:
            self._endtime = time.time() + timeout

    def _clear_process(self):
        """
        Close stdin/stdout of subprocess.

        It is required because of ResourceWarning in Python 3.
        """
        if self.process:
            if self.process.stdin:
                self.process.stdin.close()
            if self.process.stdout:
                self.process.stdout.close()

            self.process = None

        self._endtime = None

    def stop(self, sig=None):
        """
        Stop process running.

        Wait 10 seconds for the process to end, then just kill it.

        :param int sig: signal used to stop process run by executor.
            None for default.

        .. note::

            When gathering coverage for the subprocess in tests,
            you have to allow subprocesses to end gracefully.
        """
        if self.process is None:
            return

        if sig is None:
            sig = self._sig_stop

        os.killpg(self.process.pid, sig)

        def process_stopped():
            """Return True only only when self.process is not running."""
            return self.running() is False

        # set 10 seconds wait no matter what to kill the process
        self._set_timeout(10)
        try:
            self.wait_for(process_stopped)
        except TimeoutExpired:
            # at this moment, process got killed,
            pass

        self._clear_process()

    @contextmanager
    def stopped(self):
        """
        Stopping process for given context and starts it afterwards.

        Allows for easier writing resistance integration tests whenever one of
        the service fails.
        """
        if self.running():
            self.stop()
            yield
            self.start()

    def kill(self, wait=True, sig=None):
        """
        Kill the process if running.

        :param bool wait: set to `True` to wait for the process to end,
            or False, to simply proceed after sending signal.
        :param int sig: signal used to kill process run by  executor.
            None for default.
        """
        if sig is None:
            sig = self._sig_kill
        if self.running():
            os.killpg(self.process.pid, sig)
            if wait:
                self.process.wait()

            self._clear_process()

    def output(self):
        """Return process output."""
        if self.process is not None:
            return self.process.stdout

    def wait_for(self, wait_for):
        """
        Wait for callback to return True.

        Simply returns if wait_for condition has been met,
        raises TimeoutExpired otherwise and kills the process.

        :param callback wait_for: callback to call
        :raises: mirakuru.exceptions.TimeoutExpired
        """
        while self.check_timeout():
            if wait_for():
                return
            time.sleep(self._sleep)

        self.kill()
        raise TimeoutExpired(
            self, timeout=self._timeout
        )

    def check_timeout(self):
        """
        Check if timeout has expired.

        Returns True if there is no timeout set or the timeout has not expired.
        Kills the process and raises TimeoutExpired exception otherwise.

        This method should be used in while loops waiting for some data.

        :returns: True if timeout expired, False if not
        :rtype: bool
        """
        return self._endtime is None or time.time() <= self._endtime

    def __repr__(self):
        """Return unambiguous executor representation."""
        command = self.command
        if len(command) > 10:
            command = command[:10] + '...'
        return '<{module}.{executor}: "{command}" {id}>'.format(
            module=self.__class__.__module__,
            executor=self.__class__.__name__,
            command=command,
            id=hex(id(self))
        )

    def __str__(self):
        """Return readable executor representation."""
        return '<{module}.{executor}: "{command}">'.format(
            module=self.__class__.__module__,
            executor=self.__class__.__name__,
            command=self.command
        )


class StartCheckExecutor(Executor):

    """ Base class for executors with a pre- and after-start checks. """

    def pre_start_check(self):
        """
        Method fired before the start of executor.

        Should be overridden in order to return boolean value if some
        process is already started.
        :rtype: bool
        """
        raise NotImplementedError

    def start(self):
        """
        Start executor with additional checks.

        Checks if previous executor isn't running then start process
        (executor) and wait until it's started.
        """
        if self.pre_start_check():
            # Executor or other process is running with same config.
            raise AlreadyRunning(self)

        super(StartCheckExecutor, self).start()
        self.wait_for(self.after_start_check)

    def after_start_check(self):
        """
        Method fired after the start of executor.

        Should be overridden in order to return boolean value if executor
        can be treated as started.
        :rtype: bool
        """
        raise NotImplementedError



# Copyright (C) 2014 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS).

# This file is part of mirakuru.

# mirakuru is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# mirakuru is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with mirakuru.  If not, see <http://www.gnu.org/licenses/>.
"""Mirakuru's installation module."""

import re
import os
from setuptools import setup, find_packages


here = os.path.dirname(__file__)
with open(os.path.join(here, 'mirakuru', '__init__.py')) as v_file:
    package_version = re.compile(
        r".*__version__ = '(.*?)'", re.S).match(v_file.read()).group(1)

tests_require = ('pytest', 'pytest-cov', 'mock')
extras_require = {
    'docs': ['sphinx'],
    'tests': tests_require
}


def read(fname):
    """
    Read filename.

    :param str fname: name of a file to read
    """
    return open(os.path.join(here, fname)).read()

setup(
    name='mirakuru',
    version=package_version,
    description='Process executor for tests.',
    long_description=(
        read('README.rst')
        + '\n\n' +
        read('CHANGES.rst')
    ),
    keywords='process executor tests summon_process',
    url='https://github.com/ClearcodeHQ/mirakuru',
    author='Clearcode - The A Room',
    author_email='thearoom@clearcode.cc',
    license='LGPL',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: '
        'GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Testing',
    ],
    packages=find_packages(),
    install_requires=[],
    tests_require=tests_require,
    test_suite='tests',
    include_package_data=True,
    zip_safe=False,
    extras_require=extras_require,
)



"""Tcp executor tests."""

import pytest

from mirakuru import TCPExecutor
from mirakuru import TimeoutExpired, AlreadyRunning
from mirakuru.compat import http_server_cmd


PORT = 7986

http_server_cmd = '{0} {1}'.format(http_server_cmd, PORT)


@pytest.mark.parametrize('timeout', (None, 5))
def test_start_and_wait(timeout):
    """Test if executor await for process to accept connections."""
    command = 'bash -c "sleep 2 && nc -l 3000"'
    executor = TCPExecutor(command, 'localhost', port=3000, timeout=timeout)
    executor.start()

    assert executor.running() is True
    executor.stop()

    # check proper __str__ and __repr__ rendering:
    assert 'TCPExecutor' in repr(executor)
    assert command in str(executor)


def test_it_raises_error_on_timeout():
    """Check if TimeoutExpired gets rised correctly."""
    command = 'bash -c "sleep 10 && nc -l 3000"'
    executor = TCPExecutor(command, host='localhost', port=3000, timeout=2)

    with pytest.raises(TimeoutExpired):
        executor.start()

    assert executor.running() is False


def test_fail_if_other_executor_running():
    """Test raising AlreadyRunning exception."""
    executor = TCPExecutor(http_server_cmd, host='localhost', port=PORT)
    executor2 = TCPExecutor(http_server_cmd, host='localhost', port=PORT)

    with executor:

        assert executor.running() is True

        with pytest.raises(AlreadyRunning):
            executor2.start()

        with pytest.raises(AlreadyRunning):
            with executor2:
                pass



"""HTTP Executor tests."""
import sys
import socket
from functools import partial

import pytest

from mirakuru import HTTPExecutor
from mirakuru import TimeoutExpired, AlreadyRunning
from mirakuru.compat import HTTPConnection, OK, http_server_cmd
from tests import test_server_path


HOST = "127.0.0.1"
PORT = 7987


http_server_cmd = '{} {}'.format(http_server_cmd, PORT)
http_slow_cmd = '{python} {srv} {host}:{port}' \
    .format(python=sys.executable, srv=test_server_path, host=HOST, port=PORT)


slow_server_executor = partial(
    HTTPExecutor,
    http_slow_cmd,
    'http://{0}:{1}/'.format(HOST, PORT),
)


def connect_to_server():
    """Common test to check if can connect to server."""
    conn = HTTPConnection(HOST, PORT)
    conn.request('GET', '/')
    assert conn.getresponse().status is OK
    conn.close()


def test_executor_starts_and_waits():
    """Test if process awaits for HEAD request to be completed."""
    command = 'bash -c "sleep 3 && {0}"'.format(http_server_cmd)

    executor = HTTPExecutor(
        command,
        'http://{0}:{1}/'.format(HOST, PORT),
        timeout=20
    )
    executor.start()
    assert executor.running() is True

    connect_to_server()

    executor.stop()

    # check proper __str__ and __repr__ rendering:
    assert 'HTTPExecutor' in repr(executor)
    assert command in str(executor)


def test_shell_started_server_stops():
    """Test if executor terminates properly executor with shell=True."""
    executor = HTTPExecutor(
        http_server_cmd,
        'http://{0}:{1}/'.format(HOST, PORT),
        timeout=20,
        shell=True
    )

    with pytest.raises(socket.error):
        connect_to_server()

    with executor:
        assert executor.running() is True
        connect_to_server()

    assert executor.running() is False

    with pytest.raises(socket.error):
        connect_to_server()


def test_slow_server_starting():
    """
    Test whether or not executor awaits for slow starting servers.

    Simple example. You run gunicorn, gunicorn is working
    but you have to wait for worker procesess.
    """
    executor = slow_server_executor()
    executor.start()
    assert executor.running() is True

    connect_to_server()
    executor.stop()


def test_slow_server_timed_out():
    """Check if timeout properly expires."""
    executor = slow_server_executor(timeout=1)

    with pytest.raises(TimeoutExpired) as exc:
        executor.start()

    assert executor.running() is False
    assert 'timed out after' in str(exc)


def test_fail_if_other_executor_running():
    """Test raising AlreadyRunning exception when port is blocked."""
    executor = HTTPExecutor(
        http_server_cmd, 'http://{0}:{1}/'.format(HOST, PORT),
    )
    executor2 = HTTPExecutor(
        http_server_cmd, 'http://{0}:{1}/'.format(HOST, PORT),
    )

    with executor:

        assert executor.running() is True

        with pytest.raises(AlreadyRunning):
            executor2.start()

        with pytest.raises(AlreadyRunning) as exc:
            with executor2:
                pass
        assert 'seems to be already running' in str(exc)



"""Test basic executor functionality."""
import sys
import time
import signal
import shlex

import pytest

from mirakuru import Executor, HTTPExecutor
from mirakuru.base import StartCheckExecutor
from tests import test_server_path

sleep_300 = 'sleep 300'


@pytest.mark.parametrize('command', (sleep_300, sleep_300.split()))
def test_running_process(command):
    """Start process and shuts it down."""
    executor = Executor(command)
    executor.start()
    assert executor.running() is True
    executor.stop()
    assert executor.running() is False

    # check proper __str__ and __repr__ rendering:
    assert 'Executor' in repr(executor)
    assert 'sleep 300' in str(executor)


def test_custom_signal_stop():
    """Start process and shuts it down using signal SIGQUIT."""
    executor = Executor(sleep_300, sig_stop=signal.SIGQUIT)
    executor.start()
    assert executor.running() is True
    executor.stop()
    assert executor.running() is False


def test_stop_custom_signal_stop():
    """Start process and shuts it down using signal SIGQUIT passed to stop."""
    executor = Executor(sleep_300)
    executor.start()
    assert executor.running() is True
    executor.stop(sig=signal.SIGQUIT)
    assert executor.running() is False


def test_custom_signal_kill():
    """Start process and shuts it down using signal SIGQUIT."""
    executor = Executor(sleep_300, sig_kill=signal.SIGQUIT)
    executor.start()
    assert executor.running() is True
    executor.kill()
    assert executor.running() is False


def test_kill_custom_signal_kill():
    """Start process and shuts it down using signal SIGQUIT passed to kill."""
    executor = Executor(sleep_300)
    executor.start()
    assert executor.running() is True
    executor.kill(sig=signal.SIGQUIT)
    assert executor.running() is False


def test_running_context():
    """Start process and shuts it down."""
    executor = Executor(sleep_300)
    with executor:
        assert executor.running() is True

    assert executor.running() is False


def test_context_stopped():
    """Start for context, and shuts it for nested context."""
    executor = Executor(sleep_300)
    with executor:
        assert executor.running() is True
        with executor.stopped():
            assert executor.running() is False
        assert executor.running() is True

    assert executor.running() is False

echo_foobar = 'echo -n "foobar"'


@pytest.mark.parametrize('command', (echo_foobar, shlex.split(echo_foobar)))
def test_process_output(command):
    """Start process, check output and shut it down."""
    executor = Executor(command)
    executor.start()

    assert executor.output().read() == 'foobar'
    executor.stop()


@pytest.mark.parametrize('command', (echo_foobar, shlex.split(echo_foobar)))
def test_process_output_shell(command):
    """Start process, check output and shut it down with shell set to True."""
    executor = Executor(command, shell=True)
    executor.start()

    assert executor.output().read() == 'foobar'
    executor.stop()


def test_start_check_executor():
    """Validate StartCheckExecutor base class having NotImplemented methods."""
    executor = StartCheckExecutor(sleep_300)
    with pytest.raises(NotImplementedError):
        executor.pre_start_check()
    with pytest.raises(NotImplementedError):
        executor.after_start_check()


def test_stopping_not_yet_running_executor():
    """
    Test if Executor can be stopped even it was never running.

    We must make sure that it's possible to call .stop() and Executor will not
    raise any exception and .start() can be called afterwards.
    """
    executor = Executor(sleep_300)
    executor.stop()
    executor.start()
    assert executor.running() is True
    executor.stop()


def test_stopping_brutally():
    """
    Test if Executor is stopping insubordinate process.

    Check if the process that doesn't react to SIGTERM signal will be killed
    by executor with SIGKILL automatically.
    """
    host_port = "127.0.0.1:8000"
    cmd = '{} {} {} True'.format(sys.executable, test_server_path, host_port)
    executor = HTTPExecutor(cmd, 'http://%s/' % host_port)
    executor.start()
    assert executor.running() is True

    stop_at = time.time() + 10
    executor.stop()
    assert executor.running() is False
    assert stop_at <= time.time(), "Subprocess killed earlier than in 10 secs"
